// NOTE: vkCreateInstance is already defined in the mini_vulkan header
VULKAN_FN(vkDestroyInstance)
VULKAN_FN(vkCreateDevice)
VULKAN_FN(vkDestroyDevice)
VULKAN_FN(vkGetDeviceQueue)
VULKAN_FN(vkCreateBuffer)
VULKAN_FN(vkDestroyBuffer)
VULKAN_FN(vkDestroyCommandPool)
VULKAN_FN(vkDestroyDescriptorPool)
VULKAN_FN(vkDestroyDescriptorSetLayout)
VULKAN_FN(vkDestroyPipeline)
VULKAN_FN(vkDestroyPipelineLayout)
VULKAN_FN(vkDestroyShaderModule)
VULKAN_FN(vkEnumerateDeviceExtensionProperties)
VULKAN_FN(vkEnumeratePhysicalDevices)
VULKAN_FN(vkGetPhysicalDeviceFeatures)
VULKAN_FN(vkGetPhysicalDeviceProperties)
VULKAN_FN(vkGetPhysicalDeviceQueueFamilyProperties)
VULKAN_FN(vkGetBufferMemoryRequirements)
VULKAN_FN(vkCreateShaderModule)
VULKAN_FN(vkCreateDescriptorSetLayout)
VULKAN_FN(vkCreatePipelineLayout)
VULKAN_FN(vkCreateComputePipelines)
VULKAN_FN(vkCreateDescriptorPool)
VULKAN_FN(vkAllocateDescriptorSets)
VULKAN_FN(vkGetPhysicalDeviceMemoryProperties)
VULKAN_FN(vkAllocateMemory)
VULKAN_FN(vkBindBufferMemory)
VULKAN_FN(vkMapMemory)
VULKAN_FN(vkUnmapMemory)
VULKAN_FN(vkFreeMemory)
VULKAN_FN(vkUpdateDescriptorSets)
VULKAN_FN(vkCreateCommandPool)
VULKAN_FN(vkResetCommandPool)
VULKAN_FN(vkAllocateCommandBuffers)
VULKAN_FN(vkFreeCommandBuffers)
VULKAN_FN(vkBeginCommandBuffer)
VULKAN_FN(vkCmdBindPipeline)
VULKAN_FN(vkCmdBindDescriptorSets)
VULKAN_FN(vkCmdDispatch)
VULKAN_FN(vkQueueSubmit)
VULKAN_FN(vkQueueWaitIdle)
VULKAN_FN(vkEndCommandBuffer)
VULKAN_FN(vkCmdFillBuffer)
VULKAN_FN(vkCmdCopyBuffer)