#ifndef DAUBECHIES_CONSTANTS_H_
#define DAUBECHIES_CONSTANTS_H_

const float D0 = 0.4829629131445341f;
const float D1 = 0.83651630373780772f;
const float D2 = 0.22414386804201339f;
const float D3 = -0.12940952255126034f;

#endif  // DAUBECHIES_CONSTANTS_H_
